{##############################################################################}
{#                                                                            #}
{#  oCiGfB^ER|[lg                                          #}
{#           MxBinEditor                                                      #}
{#                                                                            #}
{#                                        VERSION: 1.00 BETA                  #}
{#                                         AUTHOR: (C)MASAHIRO SUZUKI         #}
{#                                         CREATE: 2001/12/17 -               #}
{#                                                 2002/01/18                 #}
{#                                                                            #}
{##############################################################################}

unit MxBinEditor;

interface

{$R *.res}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, 
  Grids, Math, Clipbrd, MxShareMemStream, StdCtrls;

const

  BECM_BASE               = WM_USER   + 100;
  BECM_SEARCHTEXTCHANGE   = BECM_BASE + 1;
  BECM_OPENFILE           = BECM_BASE + 2;

  crRowSelect             =  1;
  AnkKanaCode             =  [#$A1..#$DF];
  AsciiCode               =  [#$20..#$7E];
  AsciiControlCode        =  [#$00..#$1F, #$7F, #$FD..#$FF];
  SjisKanjiFirstCode      =  [#$81..#$9F, #$E0..#$FC];
  SjisKanjiSecondCode     =  [#$40..#$7E, #$80..#$FC];
  ReturnCode              =  [#$0A, #$0D];
  ByteColumn              =  19;
  BytesPerRow             =  16;
  ColumnCount             =  20;
  FirstHexColumn          =  2;
  LastHexColumn           =  17;
  ColumnOfHex             =  [FirstHexColumn..LastHexColumn];
  HexChars                =  '0123456789ABCDEF';
  DefaultInternalBuffSize =  65536 div 2;

type
  
  TMxChangeUndoEvent = procedure(Sender: TObject; UndoCount: Longint) of object;
  // ccAsciiccSJISȊO͖
  TMxCharCode        = (ccASCII, ccSJIS, ccJIS, ccEUC, ccUNICODE);
  TMxColumnBytes     = (cb1byte, cb2byte, cb4byte, cb8byte, cb16byte);
  TMxEditType        = (etModify, etInsert, etDelete, etMultiDelete,
                        etMultiInsert, etReplace, etRequestMultiUndo);

  PMxUndoRecord = ^TMxUndoRecord;
  TMxUndoRecord = packed record
    EditType: TMxEditType;
    StartPos: Longint;
    EndPos:   Longint;
    Buff:     Byte;
    pData:    PByteArray;
    DataSize: Longint;
  end;

  TMxBinEditor = class(TCustomGrid)
  private

    FSharedSearchTextStream: TMxShareMemStream;

    FCaretShowing: Boolean;
    FSavedWithoutClearUndo: Boolean;
    FPermModified: Boolean; // PI

    FFindText: String;
    FFindIgnoreCase: Boolean;

    FOnFileOpen: TNotifyEvent;
    FOnMouseWheelDown: TMouseWheelUpDownEvent;
    FOnMouseWheelUp: TMouseWheelUpDownEvent;

    FOnFileOpenAnyInstance: TNotifyEvent;
    FOnChangeSearchText: TNotifyEvent;
    FOnChangeFileSize: TNotifyEvent;
    FOnChangeReadOnlyMode: TNotifyEvent;
    FOnMoveCell: TNotifyEvent;
    FOnChangeEditMode: TNotifyEvent;
    FOnChangeUndo: TMxChangeUndoEvent;
    FUndoList: TList;
    FSaveBackup: Boolean;
    FSaveBackupCount: Integer;

    FViewUnderLine: Boolean;
    FUnderLineColor: TColor;

    FReadOnly: Boolean;
    FColumnBytes: TMxColumnBytes;
    FInternalBuffSize: Longint;
    FModifiedInternalBuff: Boolean;

    FInputHalfChar: Char;
    FModifieding: Boolean;

    FCurrentLineColor: TColor;
    FCurrentLineBGColor: TColor;
    FSelectColor: TColor;
    FSelectBGColor: TColor;

    FLeftMargin: Longint;

    FCharWidth: Integer;
    FHideOffsetHeader: Boolean;
    FHideByteView: Boolean;

    FInsertMode: Boolean;

    FMouseDown: Boolean;

    FSelecting: Boolean;
    FSelectCell1: Longint;
    FSelectCell2: Longint;

    
    FCharCode: TMxCharCode;
    FCaretBitmap: TBitmap;
    FFileSize: Longint;
    FLastCellCol: Byte;

    FInternalBuff: PByteArray;
    FInternalBuffPos: Longint;

    FTempFile: TStream;
    FTempFileHandle: THandle;
    FTempFileName: String;

    FOpenedFileName: String;
    FOffsetTextColor: TColor;
    FOffsetBGColor: TColor;
    FCursorColor: TColor;
    FClipBoardBuffSize: Longint;

    function CreateTempFile: String;
    procedure SetCellSize;
    procedure SetStreamSize;
    function Cell2Hex(ACol, ARow: Longint; var fPos: Longint; var Selected: Boolean): String;
    function ReadFromInternalBuffAtPos(APos, ACount: Longint): String;
    procedure ReadInternalBuff(APos: Longint);
    procedure MakeCaret;
    function CheckSelectCell(ACol, ARow: Longint): Boolean;
    function MousePos2CellPos(X, Y: Integer): TPoint;
    function CellPos2FilePos(ACol, ARow: Longint): Longint;
    function FilePos2CellPos(APos: Longint): TPoint;
    function GetNearestEnableCell(ACell: TPoint): TPoint;

    procedure SetCaret;
    procedure MoveCell(ACell: TPoint);
    function IsSelected(APos: Longint): Boolean;
    procedure WriteCell(ACol, ARow: Longint; ABuff: Byte; SaveUndo, DontInsert: Boolean);
    procedure WritePos(APos: Longint; ABuff: Byte; SaveUndo, DontInsert: Boolean);
    procedure InsertPos(APos: Longint; ABuff: Byte; SaveUndo: Boolean);
    procedure DeleteCell(ACol, ARow: Longint; SaveUndo: Boolean);
    procedure DeletePos(APos: Longint; SaveUndo: Boolean);
    procedure ClearUndoList;
    procedure AddUndoRecord(EditType: TMxEditType; StartPos, EndPos: Longint; ABuff: Byte; PData: PByteArray; DataSize: Longint);
    procedure FUndo(CellMove: Boolean);
    function IsLastCell(ACol, ARow: Longint): Boolean;
    procedure ShiftTempFilePos(APos, AValue: Longint);
    procedure DeleteA2B(APos, BPos: Longint; SaveUndo: Boolean);
    procedure InsertBuff(APos, ASize: Longint; pBuff: PByteArray; SaveUndo, CellMove: Boolean);
    procedure ReplaceBuff(APos, BPos, BuffSize: Longint; pBuff: PByteArray; SaveUndo, CellMove: Boolean);
    procedure CopyToClipBoard(APos, BPos: Longint);
    procedure CutToClipBoard(APos, BPos: Longint);
    procedure PasteFromClipBoard(APos: Longint);
    procedure SetInsertMode(Value: Boolean);
    procedure SetReadOnlyMode(Value: Boolean);
    function PrevCell(ACell: TPoint): TPoint;
    function NextCell(ACell: TPoint): TPoint;
    procedure SetCharCode(Value: TMxCharCode);
    procedure SetCaretColor(Value: TColor);
    procedure SetHideByteView(Value: Boolean);
    procedure SetHideOffsetHeader(Value: Boolean);
    procedure SetOffsetTextColor(Value: TColor);
    procedure SetOffsetBGColor(Value: TColor);
    procedure SetColumnBytes(Value: TMxColumnBytes);
    procedure FlashInternalBuff;
    function GetModified: Boolean;
    function SaveBackupFile(AName: String; ACount: Byte): Boolean;
    function SaveFile(fName: String): Boolean;
    procedure SetCurrentLineColor(Value: TColor);
    procedure SetCurrentLineBGColor(Value: TColor);
    procedure SetSelectColor(Value: TColor);
    procedure SetSelectBGColor(Value: TColor);
    function GetCurrentAddress: Longint;

    function FindText(AText: String; StartPos, EndPos: Longint; IgnoreCase, Prev, CellMove: Boolean): Longint;
    function Hex2Byte(AChar: Char): Integer;
    procedure SetViewUnderLine(Value: Boolean);
    procedure SetUnderLineColor(Value: TColor);
    procedure VisibleCaret(Show: Boolean);
    procedure DoPostMessage(Msg: Cardinal; wParam, lParam: Integer; ToParentWindow: Boolean);
    function GetSearchOptionsFromMemoryMappedFile: String;
//    procedure CalcScrollBars;
    function GetScrollBarPos(BarFlag: Integer): Integer;
    procedure SetScrollBarPos(BarFlag, nPos: Integer);
    procedure SetScrollBarSize;
    function AdjustTopRow(ARow: Longint): Longint;
    function AdjustLeftCol(ACol: Longint): Longint;

  protected
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState); override;
    procedure WMEraseBkgnd(var Msg:TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMSetFocus(var Msg: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Msg: TWMKillFocus); message WM_KILLFOCUS;
    function SelectCell(ACol, ARow: Longint): Boolean; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;

    procedure WMVScroll(var Msg: TWMVScroll); message WM_VSCROLL;
    procedure WMHScroll(var Msg: TWMHScroll); message WM_HSCROLL;

    procedure BECMSearchTextChange(var Msg: TMessage); message BECM_SEARCHTEXTCHANGE;
    procedure BECMOpenFile(var Msg: TMessage); message BECM_OPENFILE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
//    procedure Resize; override;
    procedure WMSize(var Msg: TWMSize); message WM_SIZE;
    procedure WMChar(var Msg: TWMChar); message WM_CHAR;
    procedure WMTimer(var Msg: TWMChar); message WM_TIMER;
    procedure CreateParams(var Params:TCreateParams); override;
    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean; override;


  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Modified: Boolean read GetModified;
    property CurrentAddress: Longint read GetCurrentAddress;
    property SearchText: String read FFindText;
    property SearchIgnoreCase: Boolean read FFindIgnoreCase;

    function LoadFromFile(fName: String): Boolean;
    function SaveToFile(fName: String; NoClearUndo: Boolean): Boolean;
    function SelectSize: Longint;
    function CanDelete: Boolean;
    procedure Undo;
    procedure Cut;
    procedure Copy;
    procedure Paste;
    procedure Delete;
    procedure InsertFile(fName: String);
    procedure InputStr(AStr: String);
    function Find(AText: String; IgnoreCase, Prev, FromTop, CellMove: Boolean): Longint;
    function FindNext: Longint;
    function FindPrev: Longint;
    function CanFindNext: Boolean;
    function JumpTo(AValue: Longint): Boolean;
    procedure SetSelection(APos, BPos: Longint);
    procedure OffsetJump;
    procedure SetSearchOptions(Text: String; IgnoreCase, NoSendMessage: Boolean);
    function GetInstanceCount: Integer;
    procedure CloseAllInstance;
    procedure ShowPrevWindow;
    procedure CopyText;
    function Replace(APos: Integer; AStr, BStr: String; All: Boolean): Integer;
    function Row2HexStr(APos, BPos, ARow: Longint): String;
    procedure GetSelectRange(var SelStart, SelEnd: Longint);
    function GetByteViewText(APos, ACount: Longint): String;


  published

    property OnFileOpen: TNotifyEvent read FOnFileOpen write FOnFileOpen;
    property OnFileOpenAnyInstance: TNotifyEvent read FOnFileOpenAnyInstance write FOnFileOpenAnyInstance;

    property OnChangeSearchText: TNotifyEvent read FOnChangeSearchText write FOnChangeSearchText;
    property OnChangeFileSize: TNotifyEvent read FOnChangeFileSize write FOnChangeFileSize;
    property OnChangeReadOnlyMode: TNotifyEvent read FOnChangeReadOnlyMode write FOnChangeReadOnlyMode;
    property OnChangeUndo: TMxChangeUndoEvent read FOnChangeUndo write FOnChangeUndo;
    property OnMoveCell: TNotifyEvent read FOnMoveCell write FOnMoveCell;
    property OnChangeEditMode: TNotifyEvent read FOnChangeEditMode write FOnChangeEditMode;

    property HideByteView: Boolean read FHideByteView write SetHideByteView;
    property HideOffsetHeader: Boolean read FHideOffsetHeader write SetHideOffsetHeader;
    property CaretColor: TColor read FCursorColor write SetCaretColor;
    property CharCode: TMxCharCode read FCharCode write SetCharCode;
    property AddressColor: TColor read FOffsetTextColor write SetOffsetTextColor;
    property AddressBGColor: TColor read FOffsetBGColor write SetOffsetBGColor;
    property CurrentLineColor: TColor read FCurrentLineColor write SetCurrentLineColor;
    property CurrentLineBGColor: TColor read FCurrentLineBGColor write SetCurrentLineBGColor;
    property SelectColor: TColor read FSelectColor write SetSelectColor;
    property SelectBGColor: TColor read FSelectBGColor write SetSelectBGColor;
    property ColumnBytes: TMxColumnBytes read FColumnBytes write SetColumnBytes;
    property FileName: String read FOpenedFileName;
    property SaveBackup: Boolean read FSaveBackup write FSaveBackup;
    property SaveBackupCount: Integer read FSaveBackupCount write FSaveBackupCount;
    property ReadOnlyMode: Boolean read FReadOnly write SetReadOnlyMode;
    property InsertMode: Boolean read FInsertMode write SetInsertMode;
    property FileSize: Longint read FFileSize;
    property ViewUnderLine: Boolean read FViewUnderLine write SetViewUnderLine;
    property UnderLineColor: TColor read FUnderLineColor write SetUnderLineColor;

    property Align;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property ParentCtl3D;
    property ParentShowHint;
    property PopupMenu;
//    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
  end;


var
  CF_MXBINARY: WORD;
  fBeHandleList: TList;
  hFirst, hChWnd: HWND;

procedure Register;

implementation


{##############################################################################}

procedure Register;
begin
  RegisterComponents('masa-s', [TMxBinEditor]);
end;

{##############################################################################}

function GetTopWindow(AHandle: HWND): HWND;
var
  ParentWnd: HWND;
begin
  Result := AHandle;
  ParentWnd := GetParent(Result);
  while ParentWnd <> 0 do
  begin
    Result := ParentWnd;
    ParentWnd := GetParent(Result);
  end;
end;

{##############################################################################}

function EnumChildProc(hChWnd: HWND; lParam: LongInt): BOOL; stdcall; export;
var
  pBuff: PChar;
  BuffSize: Longint;
  pHwnd: PLongword;
begin
  Result := True;
  EnumChildWindows(hChWnd, @EnumChildProc, 0);
  pBuff := StrAlloc(255);
  try
    BuffSize := GetClassName(hChWnd, pBuff, 255);
    if BuffSize <> 0 then
    begin
      if pBuff = 'TMxBinEditor' then
      begin
        New(pHwnd);
        pHwnd^ := hChWnd;  
        fBeHandleList.Add(pHwnd);
      end;
    end;
  finally
    StrDispose(pBuff);
  end;
end;

{##############################################################################}

function EnumWindowsProc(hWindow: HWND; lParam: LongInt): Boolean; stdcall;
begin
  Result := False;
  if hFirst = 0 then
    hFirst := hWindow
  else if hWindow = hFirst then
    Exit;

//  if IsWindowVisible(hWindow) then
    // qWindow
    EnumChildWindows(hWindow, @EnumChildProc, 0);
  Result := True;
end;

{##############################################################################}

function CreateBinEditorHandleList: Integer;
var
  Chk: Boolean;
begin
  if fBeHandleList = nil then
    fBeHandleList := TList.Create
  else
    fBeHandleList.Clear;

  hFirst:= 0;
  if EnumWindows(@EnumWindowsProc, 0) then
  begin
    Chk := True;
    while Chk do
      Chk := EnumWindows(@EnumWindowsProc, 0);
  end;


  Result := fBeHandleList.Count;
end;

{##############################################################################}

procedure DestroyBinEditorHandleList;
var
  i: Integer;
begin
  if FBeHandleList <> nil then
  begin
    for i := 0 to fBeHandleList.Count - 1 do
      Dispose(PLongword(fBeHandleList[i]));
    FreeAndNil(fBeHandleList);
  end;
end;

{##############################################################################}

function MsgBox(Msg: String; Flags: UINT): Integer;
begin
  Result := Application.MessageBox(PChar(Msg), PChar(Application.Title), Flags);
end;

{##############################################################################}

function TranslateStr(AStr: String; ACharCode: TMxCharCode; NoChangeReturn: Boolean): String;
var
  i, wLen: Integer;
  wChar: Char;
  PrevCharIsKanjiFirstByte: Boolean;
  RepChar: Char;

begin

  PrevCharIsKanjiFirstByte := False;
  RepChar := '.';

  wLen := Length(AStr);
  SetLength(Result, wLen);

  for i := 1 to wLen do
  begin
    wChar := AStr[i];  

    case ACharCode of

(**)  ccASCII:
        if (wChar in AsciiCode) or (NoChangeReturn and (wChar in ReturnCode)) then
          Result[i] := wChar
        else
          Result[i] := RepChar;

(**)  ccSJIS:
        begin
          if wChar in AsciiCode then
          begin
            Result[i] := wChar;
            if PrevCharIsKanjiFirstByte then
            begin
              if not (wChar in SjisKanjiSecondCode) then
                Result[i - 1] := RepChar;
            end;
            PrevCharIsKanjiFirstByte := False;
          end else if wChar in AsciiControlCode then
          begin
            if NoChangeReturn and (wChar in ReturnCode) then
            begin
              Result[i] := wChar;
            end else
            begin
              Result[i] := RepChar;
              if PrevCharIsKanjiFirstByte then
                Result[i - 1] := RepChar;
            end;
            PrevCharIsKanjiFirstByte := False;
          end else
          begin
            if wChar in SjisKanjiSecondCode then
            begin
              if PrevCharIsKanjiFirstByte then
              begin
                Result[i] := wChar;
                PrevCharIsKanjiFirstByte := False;
              end else if (wChar in AnkKanaCode)  then
              begin
                Result[i] := wChar;
                PrevCharIsKanjiFirstByte := False;
              end else if (wChar in SjisKanjiFirstCode)  then
              begin
                Result[i] := wChar;
                PrevCharIsKanjiFirstByte := True;
              end else
              begin
                Result[i] := RepChar;
                PrevCharIsKanjiFirstByte := False;
              end;
            end else
            begin
              // ̂P
              Result[i] := wChar;
              if PrevCharIsKanjiFirstByte then
                Result[i - 1] := RepChar;
              PrevCharIsKanjiFirstByte := True;
            end;
          end;
        end;

      else
        Result[i] := wChar

    end;
  end;
end;

{##############################################################################}

function GetTempFileName : String;
var
   pPath : String;
begin
  SetLength (pPath, MAX_PATH + 1);
  SetLength (pPath, GetTempPath(MAX_PATH, @pPath[1]));
  pPath := Trim(pPath);
  if not IsPathDelimiter(pPath, Length(pPath)) then
    pPath := pPath + '\';
  repeat
    Result := pPath + IntToHex(GetTickCount, 8) + '.tmp';
  until not FileExists(Result);
end;

{##############################################################################}

function MakeTempFile(FileName: String): Cardinal;
begin
  Result := CreateFile(PChar(FileName),
                       GENERIC_READ or GENERIC_WRITE, 0, nil, CREATE_ALWAYS,
                       FILE_ATTRIBUTE_TEMPORARY, 0);
  if Result = INVALID_HANDLE_VALUE then
    raise Exception.Create(SysErrorMessage(GetLastError));
end;

{##############################################################################}

procedure SwapInt(var AInt, BInt: Longint);
var
  wInt: Longint;
begin
  wInt := AInt;
  AInt := BInt;
  BInt := wInt;
end;

{##############################################################################}

procedure SortInt(var AInt, BInt: Longint);
begin
  if AInt > BInt then
    SwapInt(AInt, BInt);
end;

{##############################################################################}

constructor TMxBinEditor.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  ScrollBars := ssNone;
  FSharedSearchTextStream := TMxShareMemStream.Create('MX_SHAREMEM_SEARCHTEXT', 260);
  GetSearchOptionsFromMemoryMappedFile;

  TimedScroll([sdLeft]);

  Screen.Cursors[crRowSelect] := LoadCursor(HInstance, PChar('CRROWSELECT'));
  GetMem(FInternalBuff, DefaultInternalBuffSize);
  FUndoList              := TList.Create;
  FCaretBitmap           := TBitmap.Create;
  FInternalBuffSize      := DefaultInternalBuffSize;
  FInternalBuffPos       := -1;

  Col                    := 2;
  Row                    := 1;
  Font.Name              := 'FixedSys';
  Font.Size              := 14;
  BorderStyle            := bsSingle;
  DefaultDrawing         := False;
  Options                := [goTabs, goThumbTracking];
  CTL3D                  := False;
//  Cursor := crIBeam;

  Font.Color             := $000000;
  Color                  := $FFFFFF;
  FOffsetTextColor       := $984E00;
  FOffsetBGColor         := $FFFFFF;
  FCursorColor           := $FFFFFF;
  FCurrentLineColor      := $000080;
  FCurrentLineBGColor    := $FFFFFF;
  FSelectColor           := $000000;
  FSelectBGColor         := $FFEECC;
  FUnderLineColor        := $FF0000;
  ColCount               := ColumnCount;
  RowCount               := 2;
  FTempFileName          := CreateTempFile;
  FFileSize              := 0;
  FixedRows              := 1;
  FixedCols              := 2;
  FLeftMargin            := 10;
  FClipBoardBuffSize     := 0;
  FInsertMode            := False;
  FCharCode              := ccAscii;
  FSavedWithoutClearUndo := False;
  FPermModified          := False;
  FOpenedFileName        := '';
  FViewUnderLine         := False;
  FCaretShowing          := False;
  FSaveBackup            := False;
  FSaveBackupCount       := 10;
  FSelectCell1           := -1;
  FSelectCell2           := -1;

  ReadInternalBuff(0);
  SetCellSize;
  SetStreamSize;


end;

{##############################################################################}

function TMxBinEditor.CreateTempFile: String;
var
  wTempFileName: String;
begin

  if FTempFile <> nil  then
  begin
    FreeAndNil(FTempFile);
    FileClose(FTempFileHandle);
    DeleteFile(FTempFileName);
    FFileSize := 0;
    FTempFileName := '';
  end;

  try
    wTempFileName := GetTempFileName;
    FTempFileHandle := MakeTempFile(wTempFileName);
    FTempFile := THandleStream.Create(FTempFileHandle);
  except
    FileClose(FTempFileHandle);
    FreeAndNil(FTempFile);
    DeleteFile(wTempFileName);
    raise;
  end;
  Result :=  wTempFileName;
end;

{##############################################################################}

destructor TMxBinEditor.Destroy;
begin
  FTempFile.Free;
  FileClose(FTempFileHandle);
  DeleteFile(FTempFileName);
  FreeMem(FInternalBuff, FInternalBuffSize);
  FCaretBitmap.Free;
  ClearUndoList;
  FUndoList.Free;

  FSharedSearchTextStream.Free;

  inherited Destroy;
end;

{##############################################################################}

procedure TMxBinEditor.SetCellSize;
var
  wBmp: TBitmap;
  i: Integer;
begin
  wBmp := TBitmap.Create;
  try
    wBmp.Canvas.Font.Assign(Self.Font);
    DefaultRowHeight := wBmp.Canvas.TextHeight(' 0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ');
    FCharWidth       := wBmp.Canvas.TextWidth('W');
    DefaultColWidth  := FCharWidth * 3;
  finally
    wBmp.Free;
  end;

  if FHideOffsetHeader then
  begin
    ColWidths[0]   := 0;
    RowHeights[0]  := 0;
    ColWidths[FirstHexColumn - 1] := 0;
  end else
  begin
    RowHeights[0]  := DefaultRowHeight;
    ColWidths[0]   := FCharWidth * 8 + 10;
    ColWidths[ByteColumn - 1] := FCharWidth * 2;
  end;

  ColWidths[1]   := FLeftMargin;

  if FHideByteView then
    ColWidths[ByteColumn] := 0
  else
    ColWidths[ByteColumn] := FCharWidth * 16;


  for i := FirstHexColumn to LastHexColumn  do
    case FColumnBytes of
       cb2byte:  if i mod 2 = 0 then
                   ColWidths[i] := FCharWidth * 2;
       cb4byte:  if (i - 1) mod 4  <> 0 then
                   ColWidths[i] := FCharWidth * 2;
       cb8byte:  if (i - 1) mod 8  <> 0 then
                   ColWidths[i] := FCharWidth * 2;
      cb16byte:  if (i - 1) mod 16 <> 0 then
                   ColWidths[i] := FCharWidth * 2;
    end;

  SetScrollBarSize;

end;

{##############################################################################}

procedure TMxBinEditor.DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);
var
  pText: String;
  i, wCol, wPos, TextRGB, BackRGB: Longint;
  wRect, wRect2: TRect;
  wSelected: Boolean;
  {----------------------------------------------------------------------------}
  procedure DoDrawText;
  begin
    with Canvas do
    begin
      SetTextColor(Handle, TextRGB);
      SetBKColor(Handle, BackRGB);
      Font.Assign(Self.Font);
      ExtTextOut(Handle, ARect.Left, ARect.Top, ETO_CLIPPED or ETO_OPAQUE,
                 @wRect, PChar(pText), Length(pText), nil);

      if FViewUnderLine and (ARow = Row) and (FSelectCell2 = -1) then
      begin
        Pen.Color := FUnderLineColor;
        MoveTo(wRect.Left, wRect.Bottom - 1);
        LineTo(wRect.Right, wRect.Bottom - 1);
      end;

    end;
  end;
  {----------------------------------------------------------------------------}
begin
  wRect := ARect;

  if ARow = 0 then
  begin
    // 1s(HEADER)̏ꍇ

    SetTextAlign(Canvas.Handle, TA_TOP or TA_LEFT);
    TextRGB := ColorToRGB(FOffsetTextColor);
    BackRGB := ColorToRGB(FOffsetBGColor);

    if ACol = 0 then
    begin
      pText := '';
    end else
    begin
      if ACol in ColumnOfHex then
        pText := '+' + IntToHex(ACol - FirstHexColumn, 1)
      else if ACol = ByteColumn then
        pText := HexChars;
    end;

    DoDrawText;
    Exit;

  end else
  begin
    // Qsڈȍ~̏ꍇ
    if ACol = 0 then
    begin
      // AhX
      pText := IntToHex((ARow - 1) * BytesPerRow, 7) + ':';
      SetTextAlign(Canvas.Handle, TA_TOP or TA_RIGHT);
      SwapInt(ARect.Left, ARect.Right);
      TextRGB := ColorToRGB(FOffsetTextColor);
      BackRGB := ColorToRGB(FOffsetBGColor);

      DoDrawText;
      Exit;

    end else
    begin
      // oCg\
      if ACol = ByteColumn then
      begin

        case FCharCode of
          ccASCII: wCol := ACol;
          else wCol := ByteColumn;
        end;
{
        wText := ReadFromInternalBuffAtPos(CellPos2FilePos(wCol, ARow) - 3 , BytesPerRow + 4);
        pText := TranslateStr(wText, FCharCode, False);
}
        pText := GetByteViewText(CellPos2FilePos(wCol, ARow) - 3, BytesPerRow + 4);

        ARect.Left := ARect.Left - FCharWidth * 3;
        SetTextAlign(Canvas.Handle, TA_TOP or TA_LEFT);

        if ARow = Row then
        begin
          TextRGB := ColorToRGB(FCurrentLineColor);
          BackRGB := ColorToRGB(FCurrentLineBGColor);
        end else
        begin
          TextRGB := ColorToRGB(Font.Color);
          BackRGB := ColorToRGB(Color);
        end;

        wSelected := False;
        wRect2 := wRect;
        for i := FirstHexColumn to LastHexColumn do
        begin
          if IsSelected(CellPos2FilePos(i, ARow)) then
          begin
            wRect2.Left  := wRect.Left + ((i - 2) * FCharWidth);
            wSelected := True;
            Break;
          end;
        end;

        if wSelected then
        begin

          for i := LastHexColumn downto FirstHexColumn do
          begin
            if IsSelected(CellPos2FilePos(i, ARow)) then
            begin
              wRect2.Right := wRect.Right - ((BytesPerRow - i + 1) * FCharWidth);
              Break;
            end;
          end;

          DoDrawText;
          InvertRect(Canvas.Handle, wRect2);

        end else
          DoDrawText;

        if ARow = Row then
        begin
          wRect.Left  := wRect.Left + ((Col - FirstHexColumn) * FCharWidth);
          wRect.Right := wRect.Left + FCharWidth;
          Canvas.Brush.Color := FCursorColor;
          Canvas.FrameRect(wRect);
        end;

      end else if (ACol = LastHexColumn + 1) or (ACol = FirstHexColumn - 1) then
      begin
        // EMPTY CELL
        pText := '';
        BackRGB := ColorToRGB(Color);
        DoDrawText;
      end else
      begin

        // HEẌ

        if (ACol = Col) and (ARow = Row) and FCaretShowing then
          SetCaretPos(ARect.Left, ARect.Top);

        pText := Cell2Hex(ACol, ARow, wPos, wSelected);

        with Canvas do
        begin
          SetTextAlign(Handle, TA_TOP or TA_LEFT);

          if wSelected then
          begin
            TextRGB := ColorToRGB(FSelectColor);
            BackRGB := ColorToRGB(FSelectBGColor);
          end else if ARow = Row then
          begin
            TextRGB := ColorToRGB(FCurrentLineColor);
            BackRGB := ColorToRGB(FCurrentLineBGColor);
          end else
          begin
            TextRGB := ColorToRGB(Font.Color);
            BackRGB := ColorToRGB(Color);
          end;
        end;

        DoDrawText;

      end;
    end;
  end;
end;

{##############################################################################}

function TMxBinEditor.LoadFromFile(fName: String): Boolean;
var
  wTempFileName: String;
  SrcFileStream: TFileStream;
begin
  try
    wTempFileName := CreateTempFile;
    SrcFileStream := TFileStream.Create(fName, fmOpenRead or fmShareDenyWrite);
    try
      try
        FTempFile.Position := 0;
        FTempFile.CopyFrom(SrcFileStream, 0);
        FTempFile.Size := SrcFileStream.Size; 
      except
        FreeAndNil(FTempFile);
        DeleteFile(wTempFileName);
        raise;
      end;
    finally
      SrcFileStream.Free;
    end;
  except
    raise;
  end;

  ClearUndoList;


  FFileSize := FTempFile.Size;
  FTempFileName := wTempFileName;
  FOpenedFileName := fName;
  SetStreamSize;
  FInternalBuffPos := -1;
  FModifiedInternalBuff := False;
  FSavedWithoutClearUndo := False;
  FPermModified :=False;
  MoveCell(Point(2, 1));
  FSelecting := False;
  FSelectCell2 := -1;
  Invalidate;

  if Assigned(FOnFileOpen) then
    FOnFileOpen(Self);

  DoPostMessage(BECM_OPENFILE, 0, 0, False);
  SetScrollBarSize;


  if Assigned(FOnChangeFileSize) then
    FOnChangeFileSize(Self);

  Result := True;

end;

{##############################################################################}

function TMxBinEditor.SaveFile(fName: String): Boolean;
var
  DstFileStream: TFileStream;
begin
  Result := False;
  try
    FLashInternalBuff;
    DstFileStream := TFileStream.Create(fName, fmCreate or fmShareDenyWrite);
    try
      try
        FTempFile.Position := 0;
        DstFileStream.CopyFrom(FTempFile, FFileSize);
      except
        Exit;
      end;
    finally
      DstFileStream.Free;
    end;
  except
    Exit;
  end;
  Result := True;
end;

{##############################################################################}

function TMxBinEditor.SaveToFile(fName: String; NoClearUndo: Boolean): Boolean;
begin
  Result := False;

  if FSaveBackup then
  begin
    if not SaveBackupFile(fName, FSaveBackupCount) then
      if MsgBox('obNAbvt@C̍쐬Ɏs܂B' + #13#10 + 
                'obNAbv쐬Ȃŕۑ𑱍s܂H',
                MB_YESNO + MB_ICONWARNING) <> IDNO then Exit;
  end;

  Result := SaveFile(fName);

  if Result then
  begin
    FOpenedFileName := fName;
    if not NoClearUndo then
    begin
      ClearUndoList;
      FPermModified :=False;
      FSavedWithoutClearUndo := False;
    end else
    begin
      FSavedWithoutClearUndo := True;
      FPermModified :=False;

      if Assigned(FOnChangeUndo) then
        FOnChangeUndo(Self, FUndoList.Count);
    end;
  end;

end;

{##############################################################################}

procedure TMxBinEditor.SetStreamSize;
begin
  if FFileSize > 0 then
  begin
    RowCount := Max(2, FFileSize div BytesPerRow + 2);
    FLastCellCol := (FFileSize mod BytesPerRow) + FirstHexColumn;
  end else
  begin
    RowCount := 2;
    FLastCellCol := FirstHexColumn;
  end;

  SetScrollBarSize;

  if Assigned(FOnChangeFileSize) then
    FOnChangeFileSize(Self);

end;

{##############################################################################}

function TMxBinEditor.Cell2Hex(ACol, ARow: Longint; var fPos: Longint; var Selected: Boolean): String;
var
  wkBuff: Byte;
begin
  Result := '';

  if FFileSize = 0 then
  begin
    fPos := -1;
    Selected := False;
    Exit;
  end;

  if ACol in ColumnOfHex then
  begin
    fPos := CellPos2FilePos(ACol, ARow);
    Selected := IsSelected(fPos);
    if (fPos >= 0) and (fPos < FFileSize) then
    begin
      ReadInternalBuff(fPos);
      wkBuff := FInternalBuff[fPos - FInternalBuffPos];
      Result := HexChars[wkBuff shr 4 + 1] + HexChars[wkBuff and 15 + 1];
    end;
  end;

end;

{##############################################################################}

function TMxBinEditor.ReadFromInternalBuffAtPos(APos, ACount: Longint): String;
var
  wCount, i: Longint;
begin
  if FFileSize = 0 then Exit;

  if APos + ACount >= FFileSize then
    wCount := FFileSize - APos
  else
    wCount := ACount;

  SetLength(Result, wCount);

  for i := 1 to wCount do
  begin
    ReadInternalBuff(APos + i - 1);
    Result[i] := Char(FInternalBuff[APos - FInternalBuffPos + i - 1]);
  end;

end;

{##############################################################################}

procedure TMxBinEditor.WMEraseBkgnd(var Msg:TWMEraseBkgnd);
begin
//
end;

{##############################################################################}

procedure TMxBinEditor.WMSetFocus(var Msg: TWMSetFocus);
begin
  inherited;
  MakeCaret;
  SetCaretPos(-100, -100);
  VisibleCaret(True);
end;

{##############################################################################}

procedure TMxBinEditor.WMKillFocus(var Msg: TWMKillFocus);
begin
  VisibleCaret(False);
  if DestroyCaret = False then
    MsgBox(SysErrorMessage(GetLastError), 0);
  inherited;
end;

{##############################################################################}

procedure TMxBinEditor.FlashInternalBuff;
var
  fPos: Longint;
  wFileHandle: THandle;
  wStream: TStream;
  wFileName: String;
begin

  if not FModifiedInternalBuff then Exit;

  // obt@gkĂꍇTempFile̓eVtgB
  if FInternalBuffSize <> DefaultInternalBuffSize then
  begin
    wFileName := GetTempFileName;
    wFileHandle := MakeTempFile(wFileName);
    try
      wStream := THandleStream.Create(wFileHandle);
      try
        wStream.Position := 0;
        wStream.Write(FInternalBuff^, FInternalBuffSize);


        fPos := FInternalBuffPos + DefaultInternalBuffSize;

        if fPos < FFileSize then
        begin
          FTempFile.Position := fPos;
          wStream.CopyFrom(FTempFile, FFileSize - fPos);
        end;

        FTempFile.Position := FInternalBuffPos;
        FTempFile.CopyFrom(wStream, 0);
        FTempFile.Size := FFileSize;

        ReallocMem(FInternalBuff, DefaultInternalBuffSize);
        FInternalBuffSize := DefaultInternalBuffSize;

      finally
        wStream.Free;
      end;
    finally
      FileClose(wFileHandle);
      DeleteFile(wFileName);
    end;
  end else
  begin
    // gkĂȂꍇ͂̂܂ܓobt@߂
    FTempFile.Position := FInternalBuffPos;
    FTempFile.Write(FInternalBuff^, Min(DefaultInternalBuffSize, FTempFile.Size - FInternalBuffPos));
  end;

  FModifiedInternalBuff := False;

end;

{##############################################################################}

procedure TMxBinEditor.ReadInternalBuff(APos: Longint);
var
  fPos: Longint;
begin

  if (FInternalBuffPos = -1 ) or (APos < FInternalBuffPos{ + BytesPerRow}) or
     ((APos + BytesPerRow) > (FInternalBuffPos + FInternalBuffSize)) then
  begin

    FlashInternalBuff;
    FModifiedInternalBuff := False;

    fPos := Max(0, aPos - (DefaultInternalBuffSize div 2));
    FTempFile.Position := fPos;
    FTempFile.Read(FInternalBuff^, DefaultInternalBuffSize);
    FInternalBuffPos := fPos;

  end;
end;

{##############################################################################}

procedure TMxBinEditor.MakeCaret;
begin
  if not HandleAllocated then Exit;

  VisibleCaret(False);

  with FCaretBitmap do
  begin
    Width  := FCharWidth;
    Height := DefaultRowHeight;

    Canvas.Brush.Color := clBlack;
    Canvas.FillRect (Rect(0,0, DefaultColWidth, DefaultRowHeight));
    Canvas.Brush.Color := FCursorColor;

    if FInsertMode and (not FReadOnly) then
      Canvas.FillRect(Rect(0, 0, 2, DefaultRowHeight))
    else
      Canvas.FillRect(Rect(0, 0, DefaultColWidth, DefaultRowHeight));

    CreateCaret(Self.Handle, Handle, 0, 0);
  end;
end;

{##############################################################################}

function TMxBinEditor.SelectCell(ACol, ARow: Longint): Boolean;
begin
  Result := CheckSelectCell(ACol, ARow);
end;

{##############################################################################}

function TMxBinEditor.CheckSelectCell(ACol, ARow: Longint): Boolean;
begin

  Result := (inherited SelectCell(ACol, ARow)) and
            ((ARow > 0) and (ARow < RowCount)) and (ACol in ColumnOfHex) and
            not ((ARow = RowCount - 1) and (ACol > FLastCellCol));

end;

{##############################################################################}

procedure TMxBinEditor.KeyDown(var Key: Word; Shift: TShiftState);
var
  wCell: TPoint;
  wChar: Char;
begin
  if (Key = VK_TAB) and (ssShift in Shift) then
    Key := VK_LEFT;


  if Key = VK_SPACE then
  begin
    if ssShift in Shift then
      Key := VK_PRIOR
    else
      Key := VK_NEXT;
  end;

  wChar := Char(Key);
  if (wChar = 'C') and (GetKeyState(VK_CONTROL) and (not 1) <> 0) then
  begin
  // COPY
    if FSelectCell2 <> - 1 then
      CopyToClipBoard(FSelectCell1, FSelectCell2);
    Exit;
  end else if (wChar = 'X') and (GetKeyState(VK_CONTROL) and (not 1) <> 0) then
  begin
  // CUT
    if FSelectCell2 <> - 1 then
      CutToClipBoard(FSelectCell1, FSelectCell2);
    Exit;
  end else if (wChar = 'V') and (GetKeyState(VK_CONTROL) and (not 1) <> 0) then
  begin
  // PASTE
    if FSelectCell1 = FFileSize - 1 then
      PasteFromClipBoard(FSelectCell1 + 1)
    else
      PasteFromClipBoard(FSelectCell1);
    Exit;
  end;

  wCell := Point(Col, Row);
  case Key of

    VK_DELETE:
      begin
        if (GetKeyState(VK_CONTROL) and (not 1) <> 0) then
          Delete;

      end;
    VK_INSERT:
      begin
        SetInsertMode(not FInsertMode);
      end;

    VK_BACK,
    VK_LEFT:
      begin
        FSelecting := False;
        wCell := PrevCell(wCell);
        MoveCell(wCell);
      end;

    VK_TAB,
    VK_RIGHT:
      begin
        FSelecting := False;
        wCell := NextCell(wCell);
        MoveCell(wCell);
      end;
    VK_UP:
      begin
        FSelecting := False;
        Dec(wCell.y);
        MoveCell(wCell);
      end;
    VK_DOWN:
      begin
        FSelecting := False;
        Inc(wCell.y);
        MoveCell(wCell);
      end;
    VK_PRIOR:
      begin
        FSelecting := False;
        Dec(wCell.y, VisibleRowCount);
        wCell.y := Max(wCell.y, 1);
        MoveCell(wCell);
      end;

    VK_NEXT:
      begin
        FSelecting := False;
        Inc(wCell.y, VisibleRowCount);
        wCell.y := Min(wCell.y, RowCount - 1);
        MoveCell(wCell);
      end;

    VK_HOME:
      begin
        FSelecting := False;
        wCell.x := FirstHexColumn;
        if GetKeyState(VK_CONTROL) and (not 1) <> 0 then
          wCell.y := 1;
        MoveCell(wCell);
      end;

    VK_END:
      begin
        FSelecting := False;
        if GetKeyState(VK_CONTROL) and (not 1) <> 0 then
        begin
          wCell.x := FLastCellCol;
          wCell.y := RowCount - 1;
        end else
        begin
          if wCell.y = RowCount - 1 then
            wCell.x := FLastCellCol
          else
            wCell.x := LastHexColumn;
        end;
        MoveCell(wCell);
      end;
//    else
//      inherited KeyDown(Key, Shift);
  end;
end;

{##############################################################################}

procedure TMxBinEditor.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited KeyUp(Key, Shift);
end;

{##############################################################################}

function TMxBinEditor.GetNearestEnableCell(ACell: TPoint): TPoint;
var
  CurPos: TPoint;
  wRect: TRect;
begin

  if ACell.y = -1 then
    Result.y := Row  
  else if ACell.y < TopRow then
    Result.y := Max(1, TopRow - 1)
  else
    Result.y := ACell.y;

  if ACell.x = -1 then
    Result.x := Col

  else if ACell.x = ByteColumn then
  begin
    GetCursorPos(CurPos);
    CurPos := ScreenToClient(CurPos);
    wRect := CellRect(ByteColumn, 0);
    Result.x := Min(FirstHexColumn + (CurPos.x - wRect.Left) div FCharWidth, LastHexColumn);

  end else if ACell.x < FirstHexColumn then
    Result.x := FirstHexColumn
  else if ACell.x > LastHexColumn then
    Result.x := LastHexColumn
  else
    Result.x := ACell.x;

  if (Result.x < 0) or (Result.x > LastHexColumn) then
    Result.x := LastHexColumn;

end;

{##############################################################################}

procedure TMxBinEditor.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  wInt: Integer;
  wCell: TPoint;
begin

  inherited;

  wInt := HitTest.y;
  if (Button = mbLeft) and not (ssDouble in Shift) and
     ((wInt <> HTVSCROLL) and (wInt <> HTHSCROLL)) then
  begin
      wCell := GetNearestEnableCell(MousePos2CellPos(X, Y));
      FMouseDown := True;
      FSelecting := False;
      MoveCell(wCell);
      SetCaret;
  end;


end;

{##############################################################################}

procedure TMxBinEditor.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  FMouseDown := False;
  FSelecting := False;
  inherited;
end;

{##############################################################################}

procedure TMxBinEditor.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
{
  if Y <= DefaultRowHeight then
    Cursor := crDefault
  else if X <= ColWidths[0] then
    Cursor := crRowSelect
  else
    Cursor := crIBeam;
}
  if FMouseDown and (FFileSize > 0) then
  begin
    if not FSelecting then
      FSelecting := True;

    MoveCell(GetNearestEnableCell(MousePos2CellPos(X, Y)));
    SetCaret;
  end;
end;

{##############################################################################}

function TMxBinEditor.MousePos2CellPos(X, Y: Integer): TPoint;
var
  GridCoord: TGridCoord;
begin
  GridCoord := MouseCoord(X, Y);
  Result.x := GridCoord.x;
  Result.y := GridCoord.y;
end;

{##############################################################################}

function TMxBinEditor.CellPos2FilePos(ACol, ARow: Longint): Longint;
begin
  if ACol = ByteColumn then
    Result := (ACol - ByteColumn) + ((ARow - 1) * BytesPerRow)
  else
    Result := (ACol - FirstHexColumn) + ((ARow - 1) * BytesPerRow);
end;

{##############################################################################}

function TMxBinEditor.FilePos2CellPos(APos: Longint): TPoint;
begin
  Result.x := (APos mod BytesPerRow) + FirstHexColumn;
  Result.y := APos div BytesPerRow + 1;
end;


{##############################################################################}

function TMxBinEditor.GetScrollBarPos(BarFlag: Integer): Integer;
var
  ScrollInfo: TScrollInfo;
begin
  ZeroMemory(@ScrollInfo, SizeOf(TScrollInfo));
  ScrollInfo.cbSize := SizeOf(TScrollInfo);
  ScrollInfo.fMask  := SIF_TRACKPOS;
  GetScrollInfo(Handle, BarFlag, ScrollInfo);
  Result := ScrollInfo.nTrackPos;
end;

{##############################################################################}

procedure TMxBinEditor.SetScrollBarPos(BarFlag, nPos: Integer);
var
  ScrollInfo: TScrollInfo;
begin
  ScrollInfo.cbSize := SizeOf(TScrollInfo);
  ScrollInfo.fMask  := SIF_POS;
  ScrollInfo.nPos   := nPos;
  SetScrollInfo(Handle, BarFlag, ScrollInfo, True);
end;
{##############################################################################}

procedure TMxBinEditor.WMVScroll(var Msg: TWMVScroll);
var
  NewPos: Longint;
begin
  case Msg.ScrollCode of
    SB_LINEUP        : NewPos := TopRow - 1;
    SB_LINEDOWN      : NewPos := TopRow + 1;
    SB_PAGEUP        : NewPos := TopRow - VisibleRowCount;
    SB_PAGEDOWN      : NewPos := TopRow + VisibleRowCount;
    SB_THUMBPOSITION : NewPos := GetScrollBarPos(SB_VERT);
    SB_THUMBTRACK    : NewPos := GetScrollBarPos(SB_VERT);
    else Exit;
  end;

  TopRow := AdjustTopRow(NewPos);
  if TopRow = 1 then
    SetScrollBarSize
  else
    SetScrollBarPos(SB_VERT, NewPos);

  SetCaret;

end;

{##############################################################################}

procedure TMxBinEditor.WMHScroll(var Msg: TWMHScroll);
var
  NewPos: Longint;
begin
  case Msg.ScrollCode of
    SB_LINEUP        : NewPos := LeftCol - 1;
    SB_LINEDOWN      : NewPos := LeftCol + 1;
    SB_PAGEUP        : NewPos := LeftCol - VisibleColCount;
    SB_PAGEDOWN      : NewPos := LeftCol + VisibleColCount;
    SB_THUMBPOSITION : NewPos := GetScrollBarPos(SB_HORZ);
    SB_THUMBTRACK    : NewPos := GetScrollBarPos(SB_HORZ);
    else Exit;
  end;

  LeftCol := AdjustLeftCol(NewPos);
  if LeftCol = 2 then
    SetScrollBarSize
  else
  SetScrollBarPos(SB_HORZ, NewPos);
  SetCaret;
end;

{##############################################################################}
{
procedure TMxBinEditor.Resize;
begin
  inherited;
  SetScrollBarSize;
  SetCaret;
end;
}
{##############################################################################}

procedure TMxBinEditor.WMSize(var Msg: TWMSize);
begin
  inherited;
  SetScrollBarSize;
  SetCaret;
end;

{##############################################################################}

function TMxBinEditor.PrevCell(ACell: TPoint): TPoint;
var
  wCol, wRow: Longint;
begin
  wCol := ACell.x;
  wRow := ACell.y;

  if wCol = ByteColumn then
  begin
    Dec(wCol);
    if wCol < ByteColumn then
    begin
      wCol := ByteColumn;
      Dec(wRow);
    end;
  end else
  begin

    Dec(wCol);

    if wCol < FirstHexColumn then
    begin
      wCol := LastHexColumn;
      Dec(wRow);
    end;
  end;

  Result := Point(wCol, wRow);

end;

{##############################################################################}

function TMxBinEditor.NextCell(ACell: TPoint): TPoint;
var
  wCol, wRow: Longint;
begin
  wCol := ACell.x;
  wRow := ACell.y;

  if wCol = ByteColumn then
  begin
    Inc(wCol);
    if wCol > ByteColumn then
    begin
      wCol := ByteColumn;
      Inc(wRow);
    end;
  end else
  begin
    Inc(wCol);

    if (wCol > LastHexColumn) then
    begin
      wCol := FirstHexColumn;
      Inc(wRow);
    end;
  end;

  Result := Point(wCol, wRow);

end;

{##############################################################################}

procedure TMxBinEditor.SetCharCode(Value: TMxCharCode);
var
  wRect: TRect;
begin
  if Value <> FCharCode then
  begin
    FCharCode := Value;
    FLashInternalBuff;
    FInternalBuffPos := -1;
    ReadInternalBuff(CellPos2FilePos(Col, Row));
    wRect := BoxRect(ByteColumn, TopRow, ByteColumn, TopRow + VisibleRowCount);
    InvalidateRect(Handle, @wRect, False);
  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetCaretColor(Value: TColor);
begin
  if Value <> FCursorColor then
  begin
    FCursorColor := Value;
    VisibleCaret(False);
    DestroyCaret;
    MakeCaret;
    VisibleCaret(True);
  end;
end;

{##############################################################################}

procedure TMxBinEditor.CMFontChanged(var Message: TMessage);
begin
  inherited;
  if HandleAllocated then
  begin
    SetCellSize;
    if Focused then
    begin
      VisibleCaret(False);
      DestroyCaret;
      MakeCaret;
      VisibleCaret(True);
    end;
  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetHideByteView(Value: Boolean);
begin
  if Value <> FHideByteView then
  begin
    FHideByteView := Value;
    SetCellSize;
    if Col = ByteColumn then
      MoveCell(Point(FirstHexColumn, Row));
    Invalidate;
  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetHideOffsetHeader(Value: Boolean);
begin
  if Value <> FHideOffsetHeader then
  begin
    FHideOffsetHeader := Value;
    SetCellSize;
    MoveCell(Point(Col, Row));
    SetCaret;
  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetCaret;
var
  wRect: TRect;
begin
  if not Focused then Exit;

  wRect := CellRect(Col, Row);
  if wRect.Left + wRect.Bottom = 0 then
    VisibleCaret(False)
  else
  begin
    SetCaretPos(wRect.Left, wRect.Top);
    VisibleCaret(True);
  end;

  wRect := BoxRect(0{FirstHexColumn}, TopRow, ByteColumn, TopRow + VisibleRowCount);
  InvalidateRect(Handle, @wRect, False);

end;

{##############################################################################}

procedure TMxBinEditor.MoveCell(ACell: TPoint);
var
  wCol, wRow: Longint;
  wRect: TRect;
  OrgSelPos2: Longint;
begin

  FModifieding := False;

  if ACell.y >= RowCount then
    ACell.y := RowCount - 1
  else if ACell.y < 1 then
  begin
    ACell.y := 1;
    if ACell.x = LastHexColumn then
      ACell.x := FirstHexColumn;
  end;

  if IsLastCell(ACell.x, ACell.y) then
    ACell.x := FLastCellCol;

  wCol := ACell.x;
  wRow := ACell.y;


  FSelecting := FSelecting or
                (
                 (GetKeyState(VK_SHIFT) and (not 1) <> 0) and
                 (GetKeyState(VK_TAB)   and (not 1)  = 0) and
                 (GetKeyState(VK_SPACE) and (not 1)  = 0) 
                );

  OrgSelPos2 := FSelectCell2;

  if FSelecting then
  begin
    if (wRow = RowCount - 1) and (wCol >= FLastCellCol - 1) then
    begin
      wCol := FLastCellCol - 1;
      if wCol < FirstHexColumn then
      begin
        wCol := LastHexColumn;
        Dec(wRow);
      end;
    end;
    FSelectCell2 := Min(CellPos2FilePos(wCol, wRow), FFileSize - 1);
  end else
  begin
    FSelectCell1 := Min(CellPos2FilePos(wCol, wRow), FFileSize - 1);
    FSelectCell2 := -1;
  end;

  if (ACell.x <> Col) or (ACell.y <> Row) then
  begin
    MoveColRow(ACell.x, ACell.y, True, True);
    SetCaret;
  end else if (OrgSelPos2 > -1) and (FSelectCell2 = -1) then
  begin
    wRect := BoxRect(FirstHexColumn, TopRow, ByteColumn, TopRow + VisibleRowCount);
    InvalidateRect(Handle, @wRect, False);
  end;

  SetScrollBarSize;

  if Assigned(FOnMoveCell) then
    FOnMoveCell(Self);

end;

{##############################################################################}

procedure TMxBinEditor.SetOffsetTextColor(Value: TColor);
begin
  if Value <> FOffsetTextColor then
  begin
    FOffsetTextColor := Value;
    Invalidate;
  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetOffsetBGColor(Value: TColor);
begin
  if Value <> FOffsetBGColor then
  begin
    FOffsetBGColor := Value;
    Invalidate;
  end;
end;

{##############################################################################}

function TMxBinEditor.IsSelected(APos: Longint): Boolean;
begin
  Result := (FSelectCell2 <> -1) and {(FSelectCell1 <> -1) and} 
            (((APos >= FSelectCell1) and (APos <= FSelectCell2)) or
             ((APos <= FSelectCell1) and (APos >= FSelectCell2)));
end;


{##############################################################################}

procedure TMxBinEditor.WMChar(var Msg: TWMChar);
var
  wChar: Char;
  wRect: TRect;
  wPos1, wPos2: Longint;
  wBuff: Byte;
begin

  wChar := Char(Msg.CharCode);
  if Assigned (OnKeyPress) then
    OnKeyPress (Self, wChar);


  if FReadOnly then Exit;

  wChar := StrUpper(@wChar)^;

  if not (wChar in ['0'..'9','A'..'F']) then Exit;


  if FModifieding then
  begin
    // ͊
    FSelecting := False;
    WriteCell(Col, Row, Hex2Byte(FInputHalfChar) * $10 + Hex2Byte(wChar), False, False);
    FModifieding := False;
    MoveCell(NextCell(Point(Col, Row)));
  end else
  begin
    // P
    FInputHalfChar := wChar;

    if SelectSize > 0 then
    begin
      wPos1 := FSelectCell1;
      wPos2 := FSelectCell2;
      SortInt(wPos1, wPos2);
      FSelecting := False;
      wBuff := Hex2Byte(wChar);
      ReplaceBuff(wPos1, wPos2, 1, @wBuff, True, True);
      MoveCell(FilePos2CellPos(wPos1));
    end else
    begin
      FSelecting := False;
      WriteCell(Col, Row, Hex2Byte(wChar), True, False);
    end;

    if FInsertMode then
    begin
      wRect := BoxRect(FirstHexColumn, TopRow, ByteColumn, TopRow + VisibleRowCount);
      InvalidateRect(Handle, @wRect, False);
    end;

    FModifieding := True;

  end;

  InvalidateCell(Col, Row);
  InvalidateCell(ByteColumn, Row);

end;

{##############################################################################}

procedure TMxBinEditor.WriteCell(ACol, ARow: Longint; ABuff: Byte; SaveUndo, DontInsert: Boolean);
var
  fPos: Longint;
begin
  if not (ACol in ColumnOfHex) or (ARow < 1) then Exit;

  fPos := CellPos2FilePos(ACol, ARow);
  if (fPos < 0) or (fPos > FFileSize) then Exit;

  if not DontInsert and (FInsertMode and (not FModifieding)) or (fPos = FFileSize) then
  begin
    InsertPos(fPos, ABuff, SaveUndo);
  end else
    WritePos(fPos, ABuff, SaveUndo, DontInsert);

end;

{##############################################################################}

procedure TMxBinEditor.WritePos(APos: Longint; ABuff: Byte; SaveUndo, DontInsert: Boolean);
begin
  if (APos >= 0) and (APos <= FFileSize) then
  begin

    FSelecting := False;

    if SaveUndo then
      AddUndoRecord(etModify, APos, APos, FInternalBuff[APos - FInternalBuffPos], nil, 0);

    FInternalBuff[APos - FInternalBuffPos] := ABuff;
    FModifiedInternalBuff := True;

  end;
end;

{##############################################################################}

procedure TMxBinEditor.InsertPos(APos: Longint; ABuff: Byte; SaveUndo: Boolean);
var
  bPos, i: Longint;
begin
  if (APos >= 0) and (APos <= FFileSize) then
  begin

    ReadInternalBuff(APos);

    FSelecting := False;
    
    bPos := APos - FInternalBuffPos;
    ReallocMem(FInternalBuff, FInternalBuffSize + 1);
    FInternalBuffSize := FInternalBuffSize + 1;

    for i := FInternalBuffSize - 1 downto bPos + 1 do
      FInternalBuff[i] := FInternalBuff[i - 1];

    FInternalBuff[bPos] := ABuff;
    FModifiedInternalBuff := True;

    if SaveUndo then
      AddUndoRecord(etInsert, APos, APos, 0, nil, 0);

    FTempFile.Size := FTempFile.Size + 1;
    FFileSize := FFileSize + 1;

    SetStreamSize;
  end;
end;

{##############################################################################}

procedure TMxBinEditor.DeleteCell(ACol, ARow: Longint; SaveUndo: Boolean);
begin
  if not (ACol in ColumnOfHex) or (ARow < 1) then Exit;

  DeletePos(CellPos2FilePos(ACol, ARow), SaveUndo);

end;

{##############################################################################}

procedure TMxBinEditor.DeletePos(APos: Longint; SaveUndo: Boolean);
var
  wBuff: Byte;
  bPos, i: Longint;
begin
  if (APos >= 0) and (APos <= FFileSize) then
  begin

    ReadInternalBuff(APos);
    wBuff := FInternalBuff[APos - FInternalBuffPos];

    bPos := APos - FInternalBuffPos;

    if SaveUndo then
      AddUndoRecord(etDelete, APos, APos, wBuff, nil, 0);

    for i := bPos to FInternalBuffSize - 2 do
      FInternalBuff[i] := FInternalBuff[i + 1];

    ReallocMem(FInternalBuff, FInternalBuffSize - 1);
    FInternalBuffSize := FInternalBuffSize - 1;
    FModifiedInternalBuff := True;

    FTempFile.Size := FTempFile.Size - 1;
    FFileSize := FFileSize - 1;
    SetStreamSize;

  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetColumnBytes(Value: TMxColumnBytes);
begin
  if FColumnBytes <> Value then
  begin
    FColumnBytes := Value;
    SetCellSize;
  end;
end;

{##############################################################################}

function TMxBinEditor.GetModified: Boolean;
begin
  Result := ((FUndoList.Count > 0) and (not FSavedWithoutClearUndo)) or FPermModified = True;
end;

{##############################################################################}

procedure TMxBinEditor.ClearUndoList;
var
  i: Integer;
begin
  for i := FUndoList.Count - 1 downto 0 do
  begin
    with PMxUndoRecord(FUndoList[i])^ do
      FreeMem(pData, DataSize);
    Dispose(PMxUndoRecord(FUndoList[i]));
    FUndoList.Delete(i);
  end;

  if Assigned(FOnChangeUndo) then
    FOnChangeUndo(Self, FUndoList.Count);

end;

{##############################################################################}

procedure TMxBinEditor.AddUndoRecord(EditType: TMxEditType; StartPos, EndPos: Longint; ABuff: Byte; PData: PByteArray; DataSize: Longint);
var
  pNewUndoRec: PMxUndoRecord;
begin

  pNewUndoRec := New(PMxUndoRecord);
  pNewUndoRec^.EditType  := EditType;
  pNewUndoRec^.StartPos  := StartPos;
  pNewUndoRec^.EndPos    := EndPos;
  pNewUndoRec^.Buff      := ABuff;
  pNewUndoRec^.pData     := PData;
  pNewUndoRec^.DataSize  := DataSize;

  FUndoList.Add(pNewUndoRec);

  if Assigned(FOnChangeUndo) then
    FOnChangeUndo(Self, FUndoList.Count);

end;

{##############################################################################}

procedure TMxBinEditor.FUndo(CellMove: Boolean);
var
  ReqMultiUndo: Boolean;
  pUndoRec: PMxUndoRecord;
  wCell: TPoint;
  i, BuffSize, MultiUndoCount: Integer;
begin
  if FUndoList.Count = 0 then exit;

  pUndoRec := FUndoList[FUndoList.Count - 1];

  FSelecting := False;
  FSelectCell2 := -1;

  wCell := FilePos2CellPos(pUndoRec^.StartPos);

  if CellMove then
  begin
    MoveCell(wCell);
    SetCaret;
  end;

  ReadInternalBuff(pUndoRec^.StartPos);

  ReqMultiUndo := False;
  MultiUndoCount := 0;
  
  case pUndoRec^.EditType of
    etModify: WritePos(pUndoRec^.StartPos, pUndoRec^.Buff, False, True);
    etInsert: begin
                DeletePos(pUndoRec^.StartPos, False);
              end;
    etDelete: begin
                InsertPos(pUndoRec^.StartPos, pUndoRec^.Buff, False);
              end;
    etMultiDelete:
      begin
        // ꊇ폜ꂽ̂ꊇ
        BuffSize := pUndoRec^.EndPos - pUndoRec^.StartPos + 1;
        InsertBuff(pUndoRec^.StartPos, BuffSize, pUndoRec^.pData, False, CellMove);
        FreeMem(pUndoRec^.pData, BuffSize);
      end;
    etMultiInsert:
      begin
        // ꊇ}ꂽ̂ꊇ폜
        DeleteA2B(pUndoRec^.StartPos, pUndoRec^.EndPos, False);
      end;
    etReplace:
      begin
        ReplaceBuff(pUndoRec^.StartPos, pUndoRec^.EndPos, pUndoRec^.DataSize, pUndoRec^.pData, False, CellMove);
      end;
    etRequestMultiUndo:
      begin
        ReqMultiUndo := True;
        MultiUndoCount := pUndoRec^.DataSize;
      end;
  end;

  Dispose(pUndoRec);
  FUndoList.Delete(FUndoList.Count - 1);

  if ReqMultiUndo then
    for i := 1 to MultiUndoCount do
      FUndo(False);

  if FSavedWithoutClearUndo then
    FPermModified := True;

  if Assigned(FOnChangeUndo) then
    FOnChangeUndo(Self, FUndoList.Count);
  
end;

{##############################################################################}

procedure TMxBinEditor.Undo;
begin
  FUndo(True);
end;


{##############################################################################}

function TMxBinEditor.IsLastCell(ACol, ARow: Longint): Boolean;
begin
  Result := (ARow = RowCount - 1) and (ACol = FLastCellCol);
end;

{##############################################################################}

function TMxBinEditor.SelectSize: Longint;
begin
  if FSelectCell2 = -1 then
    Result := 0
  else if FSelectCell1 > FSelectCell2 then
    Result := FSelectCell1 - FSelectCell2 + 1
  else
    Result := FSelectCell2 - FSelectCell1 + 1;
end;

{##############################################################################}
// ƃt@CAPOS̈ʒu𑊑ΓIɕωTCYAValueŎw肷
procedure TMxBinEditor.ShiftTempFilePos(APos, AValue: Longint);
var
  wFileHandle: THandle;
  wStream: TStream;
  wFileName: String;
  wValue: Longint;
begin
  if AValue = 0 then Exit;

  FLashInternalBuff;
  FInternalBuffPos := -1;

  if APos + AValue < 0 then
    wValue := -APos
  else
    wValue := AValue;

  if (FFileSize - APos > 0) then
  begin
    wFileName := GetTempFileName;
    wFileHandle := MakeTempFile(wFileName);
    try
      wStream := THandleStream.Create(wFileHandle);
      try
        wStream.Position := 0;
        FTempFile.Position := APos;

        wStream.CopyFrom(FTempFile, FFileSize - APos);

        FTempFile.Position := APos + wValue;
        FTempFile.CopyFrom(wStream, 0);

      finally
        wStream.Free;
      end;
    finally
      FileClose(wFileHandle);
      DeleteFile(wFileName);
    end;
  end;

  FTempFile.Size := FFileSize + wValue;
  FFileSize := FTempFile.Size;
  SetSTreamSize;

//  FTempFile.Position := APos;

end;

{##############################################################################}

procedure TMxBinEditor.DeleteA2B(APos, BPos: Longint; SaveUndo: Boolean);
var
  pBuff: PByteArray;
  BuffSize: Longint;
begin

  SortInt(APos, BPos);
  FlashInternalBuff;

  BuffSize := BPos - APos + 1;
  GetMem(pBuff, BuffSize);
  FTempFile.Position := APos;
  FTempFile.Read(pBuff^, BuffSize);

  ShiftTempFilePos(BPos + 1, -BuffSize);

  FSelecting := False;
  FSelectCell2 := -1;
  FInternalBuffPos := -1;

  if SaveUndo then
    AddUndoRecord(etMultiDelete, APos, BPos, 0, pBuff, 0);

  MoveCell(FilePos2CellPos(APos));
  SetCaret;

end;

{##############################################################################}

procedure TMxBinEditor.InsertBuff(APos, ASize: Longint; pBuff: PByteArray; SaveUndo, CellMove: Boolean);
begin

  if APos > FFileSize then Exit;

  FlashInternalBuff;

  FTempFile.Position := APos;
  ShiftTempFilePos(APos, ASize);
  FTempFile.Position := APos;
  FTempFile.Write(pBuff^, ASize);
  FSelecting := False;
  FSelectCell2 := -1;
  FInternalBuffPos := -1;

  if SaveUndo then
    AddUndoRecord(etMultiInsert, APos, APos + ASize - 1, 0, nil, 0);

  if CellMove then
    MoveCell(FilePos2CellPos(APos + ASize));

end;

{##############################################################################}

procedure TMxBinEditor.ReplaceBuff(APos, BPos, BuffSize: Longint; pBuff: PByteArray; SaveUndo, CellMove: Boolean);
var
  ShiftSize: Longint;
  SrcSize: Longint;
  wBuff: PByteArray;
begin
  SortInt(APos, BPos);

  if BPos >= FFileSize then
    BPos := FFileSize - 1;

  SrcSize := BPos - APos + 1;

  FlashInternalBuff;

  GetMem(wBuff, SrcSize);

  FTempFile.Position := APos;
  FTempFile.Read(wBuff^, SrcSize);
  ShiftSize := BuffSize - SrcSize;
  ShiftTempFilePos(BPos + 1, ShiftSize);
  FTempFile.Position := APos;
  FTempFile.Write(pBuff^, BuffSize);

  if SaveUndo then
    if BuffSize = 0 then
      AddUndoRecord(etMultiDelete, APos, BPos, 0, wBuff, 0)
    else
      AddUndoRecord(etReplace, APos, APos + BuffSize - 1, 0, wBuff, SrcSize);

  FSelecting := False;
  FSelectCell2 := -1;
  FInternalBuffPos := -1;

  if CellMove then
    MoveCell(FilePos2CellPos(APos + BuffSize));


end;

{##############################################################################}

procedure TMxBinEditor.CopyToClipBoard(APos, BPos: Longint);
var
  Data: Cardinal;
  DataPtr: Pointer;
  DataSize, i: Longint;
  wStr: String;
begin

  SortInt(APos, BPos);
  DataSize := BPos - APos + 1;

  FlashInternalBuff;

  Data := GlobalAlloc(GMEM_MOVEABLE, DataSize);
  try
    DataPtr := GlobalLock(Data);
    try

        FTempFile.Position := APos;
        FTempFile.Read(DataPtr^, DataSize);


        SetLength(wStr, DataSize);
        for i := 0 to DataSize - 1 do
          wStr[i + 1] := Char(PByteArray(DataPtr)^[i]);
        ClipBoard.AsText := TranslateStr(wStr, FCharCode, True);

        ClipBoard.Open;
        try
          SetClipboardData(CF_MXBINARY, Data);
        finally
          ClipBoard.Close;
        end;



    finally
      GlobalUnLock(Data);
    end;
  except
    GlobalFree(Data);
    raise;
  end;
  
end;

{##############################################################################}

procedure TMxBinEditor.CutToClipBoard(APos, BPos: Longint);
begin
  if FReadOnly then Exit;
  CopyToClipBoard(APos, BPos);
  DeleteA2B(APos, BPos, True);
end;

{##############################################################################}

procedure TMxBinEditor.PasteFromClipBoard(APos: Longint);
var
  DataSize: Longint;
  Data: Cardinal;
  DataPtr: Pointer;
  wPos1, wPos2: Longint;
begin
  if FReadOnly then Exit;

  Clipboard.Open;
  try
    Data := Clipboard.GetAsHandle(CF_MXBINARY);
    if Data = 0 then Exit;

    DataPtr := GlobalLock(Data);
    if DataPtr = nil then Exit;
    try

//      DataSize := GlobalSize(Data);
      // {̃f[^TCY擾i蔲j
      DataSize := Length(Clipboard.AsText);
      if DataSize = 0 then Exit;

      if FSelectCell2 <> - 1 then
      begin
        wPos1 := FSelectCell1;
        wPos2 := FSelectCell2;
        SortInt(wPos1, wPos2);
        ReplaceBuff(wPos1, wPos2, DataSize, DataPtr, True, True);
      end else
        InsertBuff(APos, DataSize, DataPtr, True, True);

    finally
      GlobalUnLock(Data);
    end;
  finally
    Clipboard.Close;
  end;

end;

{##############################################################################}

function TMxBinEditor.SaveBackupFile(AName: String; ACount: Byte): Boolean;
  function GetBKFileName(AName: String; ACount: Byte): String;
  var
    Ext: String;
  begin
    Ext    := ExtractFileExt(AName);
    Result := ChangeFileExt(AName, '.' +
              StringOfChar('0', 3 - Length(IntToStr(ACount))) +
              IntToStr(ACount)) + Ext;
  end;
var
  i: Byte;
  wFileName: String;
begin

  if ACount = 0 then
  begin
    Result := True;
    Exit;
  end;

  if not FileExists(AName) then
  begin
    Result := True;
    Exit;
  end;

  Result := False;

  for i := 1 to ACount do
  begin
    wFileName := GetBKFileName(AName, i);
    if FileExists(wFileName) then
      Continue
    else
    begin
      Result := RenameFile(AName, wFileName);
      Exit;
    end;
  end;

  if not Result then
  begin
    // obNAbvJEgFull̂߁At@C̃XChs
    for i := 1 to ACount - 1 do
    begin
      wFileName := GetBKFileName(AName, i);
      if FileExists(wFileName) then
        if not DeleteFile(wFileName) then
          Exit;
      if not RenameFile(GetBKFileName(AName, i + 1), wFileName) then
        Exit;
    end;

    Result := RenameFile(AName, GetBKFileName(AName, ACount));

  end;
end;

{##############################################################################}

procedure TMxBinEditor.Cut;
begin
  if FSelectCell2 = - 1 then Exit;
  CutToClipBoard(FSelectCell1, FSelectCell2);
end;

{##############################################################################}

procedure TMxBinEditor.Copy;
begin
  if FSelectCell2 = - 1 then Exit;
  CopyToClipBoard(FSelectCell1, FSelectCell2);
end;

{##############################################################################}

procedure TMxBinEditor.Paste;
begin
  if FSelectCell1 = FFileSize - 1 then
    PasteFromClipBoard(FSelectCell1 + 1)
  else
    PasteFromClipBoard(FSelectCell1);
end;

{##############################################################################}

procedure TMxBinEditor.Delete;
var
  wPos1, wPos2: Longint;
begin
  if FReadOnly then Exit;

  if FSelectCell2 <> - 1 then
  begin
    // I͈͈ꊇ폜
    wPos1 := FSelectCell1;
    wPos2 := FSelectCell2;
    SortInt(wPos1, wPos2);

    DeleteA2B(wPos1, wPos2, True);

  end else if not IsLastCell(Col, Row) then
  begin
    DeleteCell(Col, Row, True);
    SetCaret;
  end;
end;

{##############################################################################}

function TMxBinEditor.CanDelete: Boolean;
begin
  Result := not FReadOnly and ((FSelectCell2 <> - 1) or not IsLastCell(Col, Row))
end;

{##############################################################################}

procedure TMxBinEditor.SetInsertMode(Value: Boolean);
begin
  if Value <> FInsertMode then
  begin
    FInsertMode := Value;
    VisibleCaret(False);
    DestroyCaret;
    MakeCaret;
    VisibleCaret(True);
    if Assigned(FOnChangeEditMode) then
      FOnChangeEditMode(Self);
  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetReadOnlyMode(Value: Boolean);
begin
  if Value <> FReadOnly then
  begin
    FReadOnly := Value;
    VisibleCaret(False);
    DestroyCaret;
    MakeCaret;
    VisibleCaret(True);
    if Assigned(FOnChangeReadOnlyMode) then
      FOnChangeReadOnlyMode(Self);
  end;
end;

{##############################################################################}

procedure TMxBinEditor.InsertFile(fName: String);
var
  SrcStream: TFileStream;
  SelSize, wPos, wSize, wPos1, wPos2: Longint;
begin
  if FReadOnly then Exit;

  SelSize := SelectSize;
  if SelSize > 0 then
  begin
    // I͈͈ꊇ폜
    wPos1 := FSelectCell1;
    wPos2 := FSelectCell2;
    SortInt(wPos1, wPos2);
    DeleteA2B(wPos1, wPos2, True);
  end;

  SrcStream := TFileStream.Create(fName, fmOpenRead or fmShareDenyWrite);
  try
    wPos  := CellPos2FilePos(Col, Row);
    wSize := SrcStream.Size; 
    ShiftTempFilePos(wPos, wSize);
    FTempFile.Position := wPos;
    FTempFile.CopyFrom(SrcStream, 0);

    AddUndoRecord(etMultiInsert, wPos, wPos + wSize - 1, 0, nil, 0);

  finally
    SrcStream.Free;
  end;
end;

{##############################################################################}

procedure TMxBinEditor.InputStr(AStr: String);
var
  pBuff: PByteArray;
  i, BuffSize, wPos1, wPos2: Longint;
begin
  if FReadOnly then Exit;


  BuffSize := Length(AStr);
  GetMem(pBuff, BuffSize);
  try
    for i := 0 to BuffSize - 1 do
      pBuff[i] := Byte(AStr[i + 1]);

    if SelectSize > 0 then
    begin
      wPos1 := FSelectCell1;
      wPos2 := FSelectCell2;
      SortInt(wPos1, wPos2);
      ReplaceBuff(wPos1, wPos2, BuffSize, pBuff, True, True);
    end else
    begin
      wPos1 := CellPos2FilePos(Col, Row);
      wPos2 := wPos1 + BuffSize - 1;
      if FInsertMode then
        InsertBuff(wPos1, BuffSize, pBuff, True, True)
      else
        ReplaceBuff(wPos1, wPos2, BuffSize, pBuff, True, True);
    end;

  finally
    FreeMem(pBuff, BuffSize);
  end;

end;

{##############################################################################}

procedure TMxBinEditor.SetCurrentLineColor(Value: TColor);
begin
  if Value <> FCurrentLineColor then
  begin
    FCurrentLineColor := Value;
    Invalidate;
  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetCurrentLineBGColor(Value: TColor);
begin
  if Value <> FCurrentLineBGColor then
  begin
    FCurrentLineBGColor := Value;
    Invalidate;
  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetSelectColor(Value: TColor);
begin
  if Value <> FSelectColor then
  begin
    FSelectColor := Value;
    Invalidate;
  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetSelectBGColor(Value: TColor);
begin
  if Value <> FSelectBGColor then
  begin
    FSelectBGColor := Value;
    Invalidate;
  end;
end;

{##############################################################################}

function TMxBinEditor.GetCurrentAddress: Longint;
begin
  Result := CellPos2FilePos(Col, Row);
end;

{##############################################################################}

function TMxBinEditor.FindText(AText: String; StartPos, EndPos: Longint; IgnoreCase, Prev, CellMove: Boolean): Longint;
var
  pBuff: pByteArray;
  i, TextLen, MatchCount, MatchPos: Longint;
  wByte: Byte;
  {----------------------------------------------------------------------------}
  procedure LowerBuff;
  var
    i: Integer;
  begin
    for i := 0 to TextLen - 1 do
    begin
      wByte := pBuff[i];
      if Char(wByte) in AsciiCode then
      begin
        if not ((Char(wByte) in SjisKanjiSecondCode) and 
           (i > 0) and (Char(pBuff[i - 1]) in SjisKanjiFirstCode)) then
          CharLowerBuff(@Char(pBuff[i]), 1);
      end;
    end;
  end;
  {----------------------------------------------------------------------------}
begin

  Result  := -1;

  SetSearchOptions(AText, IgnoreCase, False);

  TextLen := Length(AText);
  GetMem(pBuff, TextLen);
  try
    for i := 1 to TextLen do
      pBuff[i - 1] := Byte(AText[i]);

    if IgnoreCase then
      LowerBuff;

    if Prev then
    begin

      MatchCount := TextLen;

      // FindPrev
      for i := StartPos downto EndPos do
      begin
        ReadInternalBuff(i);

        wByte := FInternalBuff[i - FInternalBuffPos];

        if IgnoreCase and (Char(wByte) in AsciiCode) then
        begin
          if not ((Char(wByte) in SjisKanjiSecondCode) and
             (i > 0) and (Char(FInternalBuff[i - FInternalBuffPos - 1]) in SjisKanjiFirstCode)) then
            CharLowerBuff(@Char(wByte), 1);
        end;

        if wByte = pBuff[MatchCount - 1] then
        begin

          Dec(MatchCount);
          if MatchCount = 0 then
          begin
            if CellMove then
            begin
              FSelecting   := True;
              FSelectCell1 := i;
              FSelectCell2 := i + TextLen - 1;
              MoveCell(FilePos2CellPos(i));
              FSelectCell1 := i;
              FSelectCell2 := i + TextLen - 1;
              SetCaret;
            end;
            Result := i;
            Break;
          end;
        end else
        begin
          MatchCount := TextLen;
        end;
      end;

    end else
    begin

      MatchPos   := 0;
      MatchCount := 0;

      // FindNext
      for i := StartPos to EndPos do
      begin
        ReadInternalBuff(i);

        wByte := FInternalBuff[i - FInternalBuffPos];

        if IgnoreCase and (Char(wByte) in AsciiCode) then
            CharLowerBuff(@Char(wByte), 1);

        if wByte = pBuff[MatchCount] then
        begin
          if MatchCount = 0 then MatchPos := i;
          Inc(MatchCount);
          if MatchCount = TextLen then
          begin
            if CellMove then
            begin
              FSelecting   := True;
              FSelectCell1 := i;
              FSelectCell2 := i + TextLen - 1;
              MoveCell(FilePos2CellPos(MatchPos));
              FSelecting   := True;
              FSelectCell1 := MatchPos;
              FSelectCell2 := i;
              SetCaret;
            end;
            Result := MatchPos;
            Break;
          end;
        end else
        begin
          MatchCount := 0;
          MatchPos   := -1;
        end;
      end;
    end;

    Invalidate;

  finally
    FreeMem(pBuff, TextLen);
  end;
end;


{##############################################################################}

function TMxBinEditor.Find(AText: String; IgnoreCase, Prev, FromTop, CellMove: Boolean): Longint;
var
  SPos, EPos: Longint;
begin
  if FromTop then
  begin
    if Prev then
      SPos := FFileSize - 1
    else
      SPos := 0;
  end else if SelectSize = 0 then
    SPos := CellPos2FilePos(Col, Row)
  else if Prev then
    SPos := Max(FSelectCell1, FSelectCell2) - 1
  else
    SPos := Min(FSelectCell1, FSelectCell2) + 1;

  if Prev then
    EPos := 0
  else
    EPos := FFileSize - 1;

  Result := FindText(AText, SPos, EPos, IgnoreCase, Prev, CellMove);
end;

{##############################################################################}

function TMxBinEditor.FindNext: Longint;
var
  fPos: Longint;
begin
  if FSelectCell2 <> -1 then
    fPos := Min(FSelectCell1, FSelectCell2) + 1
  else
    fPos := CellPos2FilePos(Col, Row);
  Result := FindText(FFindText, fPos, FFileSize - 1, FFindIgnoreCase, False, True);
end;

{##############################################################################}

function TMxBinEditor.FindPrev: Longint;
var
  fPos: Longint;
begin
  if FSelectCell2 <> -1 then
    fPos := Max(FSelectCell1, FSelectCell2) - 1
  else
    fPos := CellPos2FilePos(Col, Row);
  Result := FindText(FFindText, fPos, 0, FFindIgnoreCase, True, True);
end;

{##############################################################################}

function TMxBinEditor.CanFindNext: Boolean;
begin
  Result := FFindText <> ''; 
end;

{##############################################################################}
function TMxBinEditor.JumpTo(AValue: Longint): Boolean;
begin
  Result := False;
  if (AValue < 0) or (AValue > FFileSize) then Exit;

  MoveCell(FilePos2CellPos(AValue));
  Result := True;

end;

{##############################################################################}

procedure TMxBinEditor.SetSelection(APos, BPos: Longint);
begin
  FSelectCell1 := APos;
  FSelectCell2 := BPos;
  FSelecting   := BPos <> -1;

  if Assigned(FOnMoveCell) then
    FOnMoveCell(Self);

  Invalidate;
end;

{##############################################################################}

procedure TMxBinEditor.OffsetJump;
var
  fPos, NewPos: Longint;
begin
  fPos := CellPos2FilePos(Col, Row);
  NewPos := fPos + FInternalBuff[fPos - FInternalBuffPos];
  NewPos := NewPos + Ord(NewPos = fPos);
  if NewPos < FFileSize then
    MoveCell(FilePos2CellPos(NewPos));

end;

{##############################################################################}

function TMxBinEditor.Hex2Byte(AChar: Char): Integer;
begin
  Result := $FF; 
  case AChar of
    '0'     : Result := $0;
    '1'     : Result := $1;
    '2'     : Result := $2;
    '3'     : Result := $3;
    '4'     : Result := $4;
    '5'     : Result := $5;
    '6'     : Result := $6;
    '7'     : Result := $7;
    '8'     : Result := $8;
    '9'     : Result := $9;
    'A', 'a': Result := $A;
    'B', 'b': Result := $B;
    'C', 'c': Result := $C;
    'D', 'd': Result := $D;
    'E', 'e': Result := $E;
    'F', 'f': Result := $F;
  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetViewUnderLine(Value: Boolean);
begin
  if Value <> FViewUnderLine then
  begin
    FViewUnderLine := Value;
    InvalidateRow(Row);
  end;
end;

{##############################################################################}

procedure TMxBinEditor.SetUnderLineColor(Value: TColor);
begin
  if Value <> FUnderLineColor then
  begin
    FUnderLineColor := Value;
    InvalidateRow(Row);
  end;
end;

{##############################################################################}

procedure TMxBinEditor.VisibleCaret(Show: Boolean);
begin
  if Show then
  begin
    if not FCaretShowing then
    begin
      FCaretShowing := True;
      ShowCaret(Handle);
    end;
  end else if FCaretShowing then
  begin
    FCaretShowing := False;
    HideCaret(Handle);
  end;
end;

{##############################################################################}

procedure TMxBinEditor.BECMSearchTextChange(var Msg: TMessage);
begin
  GetSearchOptionsFromMemoryMappedFile;
end;

{##############################################################################}

procedure TMxBinEditor.BECMOpenFile(var Msg: TMessage);
begin
  if Assigned(FOnFileOpenAnyInstance) then
    FOnFileOpenAnyInstance(Self);
end;

{##############################################################################}

procedure TMxBinEditor.SetSearchOptions(Text: String; IgnoreCase, NoSendMessage: Boolean);
var
  pBuff: PChar;
  BuffSize: Byte;
begin
  if (IgnoreCase <> FFindIgnoreCase) or (Text <> FFindText) then
  begin
    FFindText := Text;
    FFindIgnoreCase := IgnoreCase;

    BuffSize := Min(Length(Text) + 1, 255);
    pBuff := StrAlloc(BuffSize);
    try
      StrPLCopy(pBuff, Text, BuffSize);
      FSharedSearchTextStream.Position := 0;
      FSharedSearchTextStream.Write(BuffSize, SizeOf(Byte));
      FSharedSearchTextStream.Write(IgnoreCase, SizeOf(Boolean));
      FSharedSearchTextStream.Write(pBuff^, BuffSize);
    finally
      StrDispose(pBuff);
    end;

    if not NoSendMessage then
      DoPostMessage(BECM_SEARCHTEXTCHANGE, 0, 0, False);

    if Assigned(FOnChangeSearchText) then
      FOnChangeSearchText(Self);

  end;
end;

{##############################################################################}

procedure TMxBinEditor.DoPostMessage(Msg: Cardinal; wParam, lParam: Integer; ToParentWindow: Boolean);
var
  MsgTo: HWND;
  i: Integer;
begin
  CreateBinEditorHandleList;
  try
    for i := 0 to fBeHandleList.Count - 1 do
    begin
      MsgTo := PLongword(fBeHandleList[i])^;
      if ToParentWindow then
        MsgTo := GetTopWindow(MsgTo);
      PostMessage(MsgTo, Msg, wParam, lParam);
    end;
  finally
    DestroyBinEditorHandleList;
  end;
end;

{##############################################################################}

function TMxBinEditor.GetSearchOptionsFromMemoryMappedFile: String;
var
  pText: PChar;
  BuffSize: Byte;
  IgnoreCase: Boolean;
begin
  try
    FSharedSearchTextStream.Position := 0;
    FSharedSearchTextStream.Read(BuffSize, SizeOf(Byte));
    FSharedSearchTextStream.Read(IgnoreCase, SizeOf(Boolean));
    FFindIgnoreCase := IgnoreCase;
    pText := StrAlloc(BuffSize);
    try
      FSharedSearchTextStream.Read(pText^, BuffSize);
      FFindText := pText;
    finally
      StrDispose(pText);
    end;
  except
    Exit;
  end;

  if Assigned(FOnChangeSearchText) then
    FOnChangeSearchText(Self);

end;

{##############################################################################}

function TMxBinEditor.GetInstanceCount: Integer;
begin
  Result := CreateBinEditorHandleList;
  DestroyBinEditorHandleList;
end;

{##############################################################################}

procedure TMxBinEditor.CloseAllInstance;
begin
  DoPostMessage(WM_SYSCOMMAND, SC_CLOSE, 0, True);
end;

{##############################################################################}

procedure TMxBinEditor.ShowPrevWindow;
var
  i, j: Integer;
begin
  j := -1;
  CreateBinEditorHandleList;
  try
    for i := 0 to fBeHandleList.Count - 1 do
    begin
      if PLongword(fBeHandleList[i])^ = Handle then
      begin
        j := i - 1;
        if j < 0 then
          j := fBeHandleList.Count - 1;  
        Break;
      end;
    end;
    if j <> -1 then
      SetForegroundWindow(GetTopWindow(THandle(fBeHandleList[j]^)));
  finally
    DestroyBinEditorHandleList;
  end;
end;

{##############################################################################}

procedure TMxBinEditor.CopyText;
var
  y, APos, BPos, ARow, BRow: Longint;
  SL: TStringList;
begin
  if SelectSize > 0 then
  begin
    APos := FSelectCell1;
    BPos := FSelectCell2;
  end else
  begin
    APos := 0;
    BPos := FFileSize;
  end;
  SortInt(APos, BPos);

  ARow := APos div BytesPerRow;
  BRow := BPos div BytesPerRow;

  SL := TStringList.Create;
  try
    for y := ARow to BRow do
      SL.Add(Row2HexStr(APos, BPos, y));
    ClipBoard.AsText := SL.Text;
  finally
    SL.Free;
  end;

end;

{##############################################################################}

procedure TMxBinEditor.WMTimer(var Msg: TWMChar);
begin
  // oCg\NbNꍇATimedScrollĂ΂Ă܂ۂ
  // 邽WM_TIMERtbNĖ悤ɂB
end;

{##############################################################################}

function TMxBinEditor.Replace(APos: Integer; AStr, BStr: String; All: Boolean): Integer;
var
  pSrc, pDst: PByteArray;
  i, SrcSize, DstSize, ShiftSize, FindPos: Longint;
begin
  Result := 0;
  SrcSize := Length(AStr);
  if SrcSize = 0 then Exit;

  DstSize := Length(BStr);
  GetMem(pDst, DstSize);
  try
    for i := 1 to DstSize do
      pDst^[i - 1] := Byte(BStr[i]);

    // uP̃t@CTCY
    ShiftSize := DstSize - SrcSize;
 
    FindPos := FindText(AStr, APos, FFileSize - 1, FFindIgnoreCase, False, False);
    while FindPos <> - 1 do
    begin
      ReadInternalBuff(FindPos);
      
      Inc(Result);
      // AhDL^
      GetMem(pSrc, SrcSize);

      for i := 0 to SrcSize - 1 do
        pSrc[i] := FInternalBuff[FindPos - FInternalBuffPos + i];
      AddUndoRecord(etReplace, FindPos, FindPos + DstSize - 1, 0, pSrc, SrcSize);

      // obt@̓eVtg
      if ShiftSize < 0 then
      begin
        // k
        for i := FindPos - FInternalBuffPos to FInternalBuffSize + ShiftSize - 1 do
          FInternalBuff[i] := FInternalBuff[i - ShiftSize];
        ReallocMem(FInternalBuff, FInternalBuffSize + ShiftSize);
      end else
      begin
        // g
        ReallocMem(FInternalBuff, FInternalBuffSize + ShiftSize);
        for i := FInternalBuffSize + ShiftSize - 1 downto FindPos - FInternalBuffPos + 1 do
          FInternalBuff[i] := FInternalBuff[i - ShiftSize];
      end;

      // u㕶
      for i := 0 to DstSize - 1 do
        FInternalBuff[FindPos - FInternalBuffPos + i] := pDst[i];
        
      FInternalBuffSize := FInternalBuffSize + ShiftSize;
      FModifiedInternalBuff := True;
      FTempFile.Size := FTempFile.Size + ShiftSize;
      FFileSize := FFileSize + ShiftSize;

      if not All then
      begin
        FSelecting   := True;
        FSelectCell1 := FindPos;
        FSelectCell2 := FindPos + DstSize - 1;
        MoveCell(FilePos2CellPos(FSelectCell2));
        FSelectCell1 := FindPos;
        FSelectCell2 := FindPos + DstSize - 1;
        SetCaret;
        Break;
      end;

      FindPos := FindText(AStr, FindPos + 1, FFileSize - 1, FFindIgnoreCase, False, False);
    end;

  finally
    FreeMem(pDst, DstSize);
  end;

  if All and (Result > 0) then
    AddUndoRecord(etRequestMultiUndo, 0, 0, 0, nil, Result);

  SetStreamSize;

end;

{##############################################################################}

procedure TMxBinEditor.CreateParams(var Params:TCreateParams);
begin
  inherited;
  with Params do //XN[o[t
    Style:=Style or WS_VSCROLL or WS_HSCROLL;
end;

{##############################################################################}

procedure TMxBinEditor.SetScrollBarSize;
var
  ScrollInfo: TScrollInfo;
  FVSBarWidth, FHSBarWidth: Integer;
begin
  if not HandleAllocated then Exit;

  ScrollInfo.cbSize := Sizeof(TScrollInfo);
  ScrollInfo.fMask  := SIF_RANGE or SIF_PAGE or SIF_POS;

  if RowCount > VisibleRowCount then
    FVSBarWidth := GetSystemMetrics(SM_CXVSCROLL)
  else
    FVSBarWidth := 0;

  if FVSBarWidth = 0 then
  begin
    ScrollInfo.nMin  := 0;
    ScrollInfo.nMax  := 0;
    ScrollInfo.nPage := 0;
    ScrollInfo.nPos  := 0;
    Invalidate;
  end else
  begin
    ScrollInfo.nMin  := 1;
    ScrollInfo.nMax  := RowCount - 1;
    ScrollInfo.nPage := VisibleRowCount;
    ScrollInfo.nPos  := TopRow;
  end;

  SetScrollInfo(Handle, SB_VERT, ScrollInfo, True);


  if (LeftCol = FirstHexColumn) and (VisibleColCount - 2 >= ByteColumn) then
    FHSBarWidth := 0
  else
    FHSBarWidth := GetSystemMetrics(SM_CXHSCROLL);

  if FHSBarWidth = 0 then
  begin
    ScrollInfo.nMin  := 0;
    ScrollInfo.nMax  := 0;
    ScrollInfo.nPage := 0;
    ScrollInfo.nPos  := 0;
  end else
  begin
    ScrollInfo.nMin  := FirstHexColumn;
    ScrollInfo.nMax  := ByteColumn;
    ScrollInfo.nPage := VisibleColCount;
    ScrollInfo.nPos  := LeftCol;
  end;

  SetScrollInfo(Handle, SB_HORZ, ScrollInfo, True);
//  ReleaseCapture;

end;


{##############################################################################}

function TMxBinEditor.AdjustTopRow(ARow: Longint): Longint;
begin
  if ARow < 1 then
    Result := 1
  else if ARow > RowCount - VisibleRowCount then
    Result := RowCount - VisibleRowCount
  else
    Result := ARow;
end;

{##############################################################################}

function TMxBinEditor.AdjustLeftCol(ACol: Longint): Longint;
begin
  if ACol < 2 then
    Result := 2
  else if ACol > ColCount - VisibleColCount then
    Result := ColCount - VisibleColCount
  else
    Result := ACol;
end;

{##############################################################################}

function TMxBinEditor.Row2HexStr(APos, BPos, ARow: Longint): String;
var
  wByte: Byte;
  x, i: Longint;
begin
  i := ARow * BytesPerRow;

  if FHideOffsetHeader then
    Result := ''
  else
    Result := IntToHex(i, 7) + ':';

  for x := FirstHexColumn to LastHexColumn do
  begin
    ReadInternalBuff(i);
    wByte := FInternalBuff[i - FInternalBuffPos];
    case FColumnBytes of
       cb1byte:  Result := Result + ' ';
       cb2byte:  if (x - 2) mod 2  = 0 then Result := Result + ' ';
       cb4byte:  if (x - 2) mod 4  = 0 then Result := Result + ' ';
       cb8byte:  if (x - 2) mod 8  = 0 then Result := Result + ' ';
      cb16byte:  if (x - 2) mod 16 = 0 then Result := Result + ' ';
    end;
    if (i < APos) or (i > BPos) then
      Result := Result + '  '
    else
      Result := Result + HexChars[wByte shr 4 + 1] + HexChars[wByte and 15 + 1];
    Inc(i);
  end;
end;

{##############################################################################}

function TMxBinEditor.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean;
begin
  if Assigned(FOnMouseWheelDown) then
    FOnMouseWheelDown(Self, Shift, MousePos, Result);
  TopRow := AdjustTopRow(TopRow + 1);
  SetCaret;
  Result := True;
end;

{##############################################################################}

function TMxBinEditor.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean;
begin
  if Assigned(FOnMouseWheelUp) then
    FOnMouseWheelUp(Self, Shift, MousePos, Result);

  TopRow := AdjustTopRow(TopRow - 1);
  SetCaret;
  Result := True;
end;

{##############################################################################}

procedure TMxBinEditor.GetSelectRange(var SelStart, SelEnd: Longint);
begin
  if SelectSize = 0 then
  begin
    SelStart := -1;
    SelEnd   := -1;
  end else
  begin
    SelStart := FSelectCell1;
    SelEnd   := FSelectCell2;
    SortInt(SelStart, SelEnd);
  end;
end;

{##############################################################################}

function TMxBinEditor.GetByteViewText(APos, ACount: Longint): String;
begin
  Result := TranslateStr(ReadFromInternalBuffAtPos(APos, ACount), FCharCode, False);
end;

{##############################################################################}

initialization
  CF_MXBINARY := RegisterClipboardFormat(PChar('CF_MXBINARY'));

{##############################################################################}

end.
