unit MxShareMemStream;

{$R-}
interface

uses SysUtils, Windows, Classes;

type
  TMxShareMemStream = class(TStream)
  protected
    FMemory: Pointer;
    FHandle: Integer;
    FSize, FPosition: Longint;
  public
    constructor Create(const Name: String; Size: Longint);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    procedure SaveToStream(Stream: TStream);
    procedure SaveToFile(const FileName: String);
    property Handle: Integer read FHandle;
    property Memory: Pointer read FMemory;
    property Size: Longint read FSize;
  end;

implementation

constructor TMxShareMemStream.Create(const Name: String; Size: Longint);
begin
  // t@C̃}bvht@C쐬
  FHandle := CreateFileMapping( $FFFFFFFF, nil, PAGE_READWRITE,
                               0, Size, PChar(Name));

  if FHandle < 0 then
    raise EFCreateError.Create('L['+Name+']J܂')
  else begin
    FMemory := MapViewOfFile(FHandle,
                  FILE_MAP_READ or FILE_MAP_WRITE, 0, 0, 0);
    FSize := Size;
    FPosition := 0;
  end;
end;

destructor TMxShareMemStream.Destroy;
begin
  if Handle >= 0 then
  begin
    UnmapViewOfFile(FMemory);
    CloseHandle(Handle);
  end;
end;

function TMxShareMemStream.Read(var Buffer; Count: Longint): Longint;
begin
  if (FPosition >= 0) and (Count >= 0) then
  begin
    Result := FSize - FPosition;
    if Result > 0 then
    begin
      if Result > Count then Result := Count;
      Move(Pointer(Longint(FMemory) + FPosition)^, Buffer, Result);
      Inc(FPosition, Result);
      Exit;
    end;
  end;
  Result := 0;
end;

function TMxShareMemStream.Write(const Buffer; Count: Longint): Longint;
var
  Pos: Longint;
begin
  if (FPosition >= 0) and (Count >= 0) then
  begin
    Pos := FPosition + Count;
    if Pos > 0 then
    begin
      if Pos > FSize then
      begin
        raise EAccessViolation.Create( 'LE');
      end;
      System.Move(Buffer, Pointer(Longint(FMemory) + FPosition)^, Count);
      FPosition := Pos;
      Result := Count;
      Exit;
    end;
  end;
  Result := 0;
end;

function TMxShareMemStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  case Origin of
    0: FPosition := Offset;
    1: Inc(FPosition, Offset);
    2: FPosition := FSize + Offset;
  end;
  Result := FPosition;
end;

procedure TMxShareMemStream.SaveToStream(Stream: TStream);
begin
  if FSize <> 0 then Stream.WriteBuffer(FMemory^, FSize);
end;

procedure TMxShareMemStream.SaveToFile(const FileName: String);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

end.
